# Artifact

Name: *Verified Lock-Free Session Channels with Linking*

## Introduction

This artifact contains a Coq formalization of Dependent Separation Protocol Channels with linking operation. Both the implementation and verification of channels are built up in the same layers as described in the paper:

| Layer                                               | Files                                                               |
| --------------------------------------------------- | ------------------------------------------------------------------- |
| Queues with a Logically Atomic Triple specification | `/theories/atomic_simple_queue.v`, `/theories/atomic_array_queue.v` |
| Queues with ghost linking                           | `/theories/atomic_lqueue.v`                                         |
| Atomic Channels                                     | `/theories/atomic_channel.v`                                        |
| Dependent Separation Protocol Channels              | `/theories/channel.v`                                               |

To show that the dependent separation protocol channel specification can be used to verify message passing programs, the artifact includes verifications of the examples described in the paper `/theories/examples`. Several examples (e.g., `/theories/examples/introduction.v`) contain closed proofs of safety (expressed using Iris' `adequacy`), stating that the example does not get stuck w.r.t. the operational semantics of [Iris's HeapLang](https://gitlab.mpi-sws.org/iris/iris/-/blob/master/docs/heap_lang.md).

This artifact uses the concurrent separation logic framework [Iris](https://gitlab.mpi-sws.org/iris/iris) and the model of dependent separation protocols defined by [Actris](https://gitlab.mpi-sws.org/iris/actris) (Hinrichsen et al. 2020).

## Hardware Dependencies

The artifact has no specfic hardware dependencies and has been tested on Linux.

## Getting Started Guide

### Building the Artifact

The artifact can be built and verified in Coq via the following steps:

#### Step 1: Create an Opam Switch

- If opam (version ≥ 2.0) is not installed, [install](https://opam.ocaml.org/doc/Install.html) it and run the following to set up opam:
      
      opam init
      eval $(opam env)

- Create and load a local switch in the artifact directory:

      opam switch create --no-install . ocaml-variants.4.14.1+options
      eval $(opam env)

#### Step 2: Install the Dependencies

- Make sure opam can find the Coq and Iris packages:

      opam repo add coq-released https://coq.inria.fr/opam/released
      opam repo add iris-dev git+https://gitlab.mpi-sws.org/iris/opam.git
      opam update

- Set the Coq version and install the artifact dependencies:

      opam pin add -n coq 8.19.1
      make builddep

- To browse the Coq sources in [CoqIDE](https://coq.inria.fr/doc/master/refman/practical-tools/coqide.html) (A graphical interactive interface for Coq), install it with

      opam install coqide

    Of course, any IDE that supports Coq can be used as well.

### Step 3: Build and Verify the Artifact

Both Coq and CoqIDE are installed in an opam switch and therefore not available globally.
In a terminal opened in the artifact directory, activate the switch using:

    eval $(opam env)

The artifact can be built using `X` cpu-cores using:

    make -j X

This verifies all the proofs in the artifact, including the examples in `/theories/examples`.

### Browsing the artifact.

The `make` command can be used to verify all proofs in the artifact. After this, individual files can be opened and stepped through in CoqIDE.

Activate the switch using `eval $(opam env)` in the artifact directory, followed by launching CoqIDE using `coqide`.

The definitions and proofs are structured as follows:

- `/theories/atomic_simple_queue.v`: The linked list implementation of queues and verification of the logically atomic triples for queues.

- `/theories/atomic_array_queue.v`: An alternative array implementation of queues and verification of the same logically atomic triples.

- `/theories/atomic_lqueue.v`: The verification of the specification of queues with ghost linking. As this layer depends only on the specification of the underlying queue, the queue implementation can be swapped out in the imports without changing the proofs.

- `/theories/atomic_channel.v`: The implementation of atomic channels as bidirectional queues, and verification of the logically atomic triple specification for such channels. This verification depends on the specification of queues with ghost linking, and therefore does not depend on a specific queue implementation.

- `/theories/channel.v`: The verification of the sequential specification of channels with dependent separation protocols and linking using for the implementation in `atomic_channel.v`.

- `/theories/proto.v`: The proof of a lemma `iProto_join` used during the proof of channel linking to merge the protocol status of two channels when linking. 

- `/theories/time_receipts.v`: A model of non-persistent time receipts ⧗ defined using an invariant and the persistent time receipts ⧖ (`steps_lb`) in Iris.

- `/theories/proofmode.v`: The Actris proof mode tactics instantiated using the channel implementation of this artifact. This provides `wp_send` and `wp_recv` tactics which are nicer to apply than applying the specifications directly.

- `/theories/examples/*.v`: The example programs defined in the paper with proofs of their sequential specification using the channel operation specification in `/theories/channel.v` and the tactics defined by `/theories/proofmode.v`.

### Key differences between artifact and paper

- In the Coq formalization, non-persistent time receipts are modeled using persistent time receipts `steps_lb` and an Iris invariant `tr_ctx`. This requires adding the `tr_ctx` to the specification similar to the ghost linking context.

- The Coq formalization hides the ghost names for the channel context using a type class. As such, we provide an additional adequacy theorem to instantiate both the type class and provide the corresponding channel context in `/theories/adequacy.v`.

  This adequacy theorem can be used to obtain a closed proof of safety (using Iris's definition of `adequacy`) w.r.t. the operational semantics. See `prog1_adequate` in `/theories/examples/introduction.v` for an example.

### Notation differences for Iris and Actris

Some notation such as for logically atomic triples and dependent separation protocols defined in Iris and Actris differ slightly between the paper and the Coq implementation. The corresponding Coq notations for these are as follows:

| Description              | Notation in Paper                     | Notation in Coq                                                             |
| ------------------------ | --------------------------------------| --------------------------------------------------------------------------- |
| Hoare Triples            | `{ P } e { v. Q }`                    | `{{{ P }}} e {{{ v. RET v; Q }}}` (*)                                       |
| Logically Atomic Triples | `<Pₚ \| x⃗. Pₐ > p < y⃗. Qₐ \| v. Qₚ>ₑ` | `Pₚ -∗ <<{ ∀∀ x₁ … xₙ, Pₐ }>> p @ e <<{ ∃∃ y₁ … yₙ, Qₐ \| v. RET v; Qₚ }>>` |
| Send protocol            | `!x⃗<v>{P}.prot`                       | `<! x₁ … xₙ> MSG v {{ P }}; prot`                                           |
| Recv protocol            | `?x⃗<v>{P}.prot`                       | `<? x₁ … xₙ> MSG v {{ P }}; prot`                                           |
| End of protocol          | `end`                                 | `END`                                                                       |
| Dual of protocol         | `p̅r̅o̅t̅`                                | `iProto_dual prot`                                                          |
| Ghost variables          | `γ ↪ᵣ v`                              | `ghost_var γ r v`                                                           |

(*) The general form of the Hoare triple in Iris is `{{{ P }}} e {{{ x₁ … xₙ. RET v; Q }}}`, where `x₁ … xₙ` are existiantial quantifiers that bind into the return value `v` and the postcondition `Q`.

For the definitions and lemmas in the paper, the correspondences between the paper and artifact are specified in the Step-by-Step instructions.

## Step By Step Instructions

### End-to-End Correctness

The channel specification proved in this artifact can be used to prove sequential specifications of programs with message passing channels. Using the adequacy theorem, closed proofs of these specifications can be used to prove safety of the programs w.r.t. the operational semantics of HeapLang.

An example of such a proof of safety is `prog1_adequate` in `/theories/examples/introduction.v`. Interpreting to the end of the print statement `Print Assumptions prog1_adequate.` prints `Closed under the global context`, indicating that the proof of `prog1_adequate` is closed (i.e., does not depend on any axioms).

### Correspondence to the Paper

The definitions in this artifact correspond to those in the paper. Below is a detailed correspondence between the definitions in the paper and those in the artifact, ordered by section in the paper.

**Introduction**

* Verification of example $prog_1$

  `/theories/examples/introduction.v`

**Section 2.1 Basic Examples**

* Verification of example programs

  `/theories/examples/basicexamples.v`

**Section 2.2 High-Level Specifications**

* Definition of channel specification

  `/theories/channel.v`

  | operation                  | lemma name                       |
  |----------------------------|----------------------------------|
  | $new\_chan$                | `new_chan_spec`                  |
  | $send$                     | `send_spec` and `send_spec_tele` |
  | $recv$                     | `recv_spec`                      |
  | $link$                     | `link_spec`                      |

**Section 2.3 Insertion Sort**

* Verification of `elem`, `empty` and `client`:

  `/theories/examples/insertionsort_all.v`

**Section 2.4 Modified Insertion Sort**

* Verification of modified `elem`, `empty` and adjusted `client'`:

  `/theories/examples/insertionsort_one.v`

**Section 3.2 Linked-List Implementation of Unidirectional Queues**

* Implementation of queue operations:

  `/theories/atomic_simple_queue.v`

* Verification of $prog_5$ example:

  `/theories/examples/queues.v`

**Section 3.3 Array-Segment Implementation of Unidirectional Queues**

* Implementation of queue operations:

  `/theories/atomic_array_queue.v`

**Section 3.4 Implementation of Session Channels**

* Implementation of channel operations:

  `/theories/atomic_channel.v`

**Section 3.5 End-to-End Correctness**

* Adequacy theorem for programs using channels with linking:

  `/theories/adequacy.v`

* Example closed proof of adequacy for $prog_1$ in the introduction:

  `/theories/introduction.v`

**Section 4.2 Invariants in Iris**

* Verification of example $prog_6$:

  `/theories/examples/queues.v`

* Example program using invariants `incr` and `prog_7`:
  
  `/theories/examples/invariants.v`

**Section 4.3 Logically-Atomic Triples in Iris**

* Logically atomic spec of `incr`:
  
  `/theories/examples/invariants.v`

**Section 4.4 Logically-Atomic Specification of Unidirectional Queues**

* Verification of logically atomic specification of linked-list and array-segment queues:

  `/theories/atomic_simple_queue.v`

  `/theories/atomic_array_queue.v`

  | operation     | lemma name        |
  |---------------|-------------------|
  | $new\_queue$  | `new_queue_spec`  |
  | $dequeue$     | `dequeue_spec`    |
  | $enqueue$     | `enqueue_spec`    |
  | $link\_queue$ | `link_queue_spec` |

* Specification and verification of $prog_8$:

  `/theories/examples/queues.v`

**Section 4.5 Verification of the Logically-Atomic Specification of Unidirectional Queues**

* Definition of queue representation predicates:

  `/theories/atomic_simple_queue.v`

  | predicate         | definition name      |
  |-------------------|----------------------|
  | $is\_queue\_list$ | `is_queue_list`      |
  | $Q$               | `is_queue_def`       |
  | $Enq$             | `enqueue_handle_def` |
  | $Deq$             | `dequeue_handle_def` |


**Section 5.1 Specification of Queues with Ghost Linking**

* Specification of ghost linking operations and rules:

  `/theories/atomic_lqueue.v`

  | operation/rule         | lemma name             |
  |------------------------|------------------------|
  | $new\_queue$           | `new_queue_lspec`      |
  | $dequeue$              | `dequeue_lspec`        |
  | $enqueue$              | `enqueue_lspec`        |
  | $make-ghost-link$      | `is_lqueue_llink`      |
  | $make-ghost-link-self$ | `is_lqueue_llink_self` |
  | $link\_queue$          | `link_queue_lspec`     |

**Section 5.2 Logically Atomic Specification of Channels**

* Representation predicates for atomic channels:

  `/theories/atomic_channel.v`

  | operation/rule | lemma name        |
  |----------------|-------------------|
  | $C$            | `is_chan_def`     |
  | $Ch$           | `chan_handle_def` |

* Specification of atomic channel operations and rules:
   
  `/theories/atomic_channel.v`
  
  | operation/rule   | lemma name       |
  |------------------|------------------|
  | $new\_chan$      | `new_chan_spec`  |
  | $send$           | `send_spec`      |
  | $recv$           | `recv_spec`      |
  | $chan-symmetric$ | `is_chan_sym`    |
  | $link$           | `link_spec` |

**Section 5.3 Verification of Queues with Ghost Linking**

* Representation Predicate Definitions:

  `/theories/atomic_lqueue.v`

  | predicate            | definition name |
  |----------------------|-----------------|
  | $lchain$             | `lchain`        |
  | $lcycle$             | `lcycle`        |
  | $Qctx_👻$            | `lqueue_ctx`    |
  | ghost queue ($Q_👻$) | `is_lqueue`     |
  | link token ($⇜$)     | `is_llink`      |

* `GHOST-CTX-ALLOC` rule: lemma `lqueue_ctx_alloc`

**Section 6.1 Protocol Linking**

Note that `prot_auth` in the paper corresponds to `proto_ctx` in Actris and this artifact. The paper name is chosen to distinguish it from the ghost linking `Qctx_👻` and channel contexts `chan_ctx`, both of which are invariant.

* Protocol linking rule `PROTO-JOIN`: lemma `iProto_join` in `/theories/proto.v`

**Section 6.2 Pairing Invariants**

The pairing invariant is defined in `/theories/pairing.v`.

* Rules for the pairing invariant:

  | rule            | lemma name          |
  |-----------------|---------------------|
  | `PAIR-ALLOC`    | `pairing_ctx_alloc` (allocates the pair invariant in an Iris invariant) |
  | `PAIR-ADD`      | `pairing_new`       |
  | `PAIR-UPDATE`   | `pairing_update`    |
  | `PAIR-UPDATE-2` | `pairing_update_2`  |


* Definitions of pairing invariant predicates:

  | predicate                 | definition name      |
  |---------------------------|----------------------|
  | $is\_pairing$             | `pairing`            |
  | $paired\_res$             | `pairing_resources`  |
  | pairing invariant ($⇄ R$) | `pairing_inv`        |
  | $Tok$                     | `is_pairing_element` |

**Section 6.3 Specification of Protocol-Based Channels**

The protocol based channels are defined in `/theories/channel.v`.

  | predicate             | definition name     |
  |-----------------------|---------------------|
  | pairing predicate $R$ | `proto_pairing`     |
  | $chan\_ctx$           | `proto_chan_ctx`    |
  | $c ↣ p$               | `iProto_mapsto_def` |

  | operation / rule           | lemma name                       |
  |----------------------------|----------------------------------|
  | $new\_chan$                | `new_chan_spec`                  |
  | $send$                     | `send_spec` and `send_spec_tele` |
  | $recv$                     | `recv_spec`                      |
  | $link$                     | `link_spec`                      |
  | allocation of  $chan\_ctx$ | `proto_chan_ctx_alloc`           |

**Section 7 Higher Order Protocols**

* The non-persistent time receipts (⧗) are modeled using Iris `steps_lb`: `/theories/time_receipts.v`

* Verification of example program $prog_9$: `/theories/examples/higherorder.v`

## Reusibility Guide

Due to the layered structure of the implementation and verification, the queue implementation can be swapped for any queue matching the same specification without altering the proofs of any higher layer (ghost linking, atomic channels, protocol channels). In fact, we provide and verified two queue implementations, one based on a linked list (`atomic_simple_queue.v`), and one based on arrays (`atomic_array_queue.v`).

The queue implementation can be swapped by changing the imported queue in `atomic_lqueue.v`, after which the artifact can be reverified for the chosen queue implementation using `make`.

Additionally, the artifact can be used to verify additional programs with message passing channels. As the top level channel specification is the similar as for [Actris](https://gitlab.mpi-sws.org/iris/actris) with the addition of the `link` operation, we include the Actris proof mode (`proofmode.v`) which provides nicer tactics for sending and receiving, rather than applying the specifications directly. We provide proofs for the example programs in the paper in `theories/examples` which may be used as reference for proving additional programs.

Note that our channel specification differs slightly from Actris' as follows: (1) the `start_chan` operation in our implementation and prior Actris versions was renamed to `fork_chan`, (2) our `new_chan` and `start_chan` operations require an additional channel context `proto_chan_ctx` which must be passed around, (3) due to the context we use `chanGS` instead of `chanG` in the implicit context, and (4) we include a `link` operation.

